/*++
/* NAME
/*	dscan 1
/* SUMMARY
/*	distributed port scanner
/* SYNOPSYS
/* .fi
/*	\fBdscan\fR [\fB-hvuRS\fR] [\fB-p \fIsec\fR] [\fB-l \fIfile\fR]
/*	\fIhost\fR \fIstart_port\fR \fIstop_port\fR \fIdhosts...\fR
/* DESCRIPTION
/*	The \fBdscan\fR command conducts a distributed port scan using a
/*	number of distributed hosts.  The folowing options are implemented:
/*
/* .IP \fB-h\fR
/*	Display the correct program usage.
/* .IP \fB-v\fR
/*	Display the program version.
/* .IP \fB-u\fR
/*	Scan UDP ports, default is TCP.
/* .IP \fB-R\fR
/*	Make the distributed hosts resolve the hostname.
/* .IP \fB-S\fR
/*	Suppress the output.
/* .IP "\fB-p \fIsec\fR"
/*	Pause for \fIsec\fR seconds before scanning the next port.
/* .IP "\fB-l \fIfile\fR"
/*	Log the scan results to \fIfile\fR.
/* .IP \fBhost\fB
/*	The host to scan.
/* .IP \fBstart_port\fB
/*	Port to start scanning from.
/* .IP \fBstop_port\fB
/*	Port to stop scanning at.
/* .IP \fBdhosts...\fB
/*	List of distributed hosts to use for the scan.
/* SEE ALSO
/*	dscand(8) distributed port scaner daemon
/*	inetd(8) internet super server
/* LICENSE
/* .ad
/* .fi
/*	The dscan license must be distributed with this software.
/* AUTHOR(S)
/*	Intrinsic <intrinsic@hushmail.com>, Ni9 Technologies
/*--*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <dscan.h>
#include <linked_list.h>
#include <network.h>

void	usage(char *);

int	main(int argc, char **argv)
{
	int	arg, suppress, i;
	unsigned int sleep_time;
	unsigned long ip_address;
	char	*log_path, packet_buffer[LENGTH_PACKET];
	struct sockaddr_in sock_info;
	struct dhost_list_s list;
	struct dhost_s *current_dhost;
	struct dscanp_packet_s packet, recv_packet;
	FILE	*log_fd;

	/*
	 * Set any variables that may change due to program arguments to their
	 * default values before reading and interpreting the arguments.
	 */
	suppress = 0;
	sleep_time = 0;
	log_path = NULL;
	packet.resolved = PACKET_RESOLVED_YES;
	packet.protocol = PACKET_PROTOCOL_TCP;
	packet.scan = PACKET_SCAN_CONNECT;

	while((arg = getopt(argc, argv, "hvuRSp:l:")) != EOF) {
		switch(arg) {
			case 'h':
				usage(argv[0]);
				exit(0);
			case 'v':
				fprintf(stderr, "%s\n", VERSION);
				exit(0);
			case 'u':
				packet.protocol = PACKET_PROTOCOL_UDP;
				break;
			case 'R':
				packet.resolved = PACKET_RESOLVED_NO;
				break;
			case 'S':
				suppress = 1;
				break;
			case 'p':
				sleep_time = atoi(optarg);
				break;
			case 'l':
				log_path = optarg;
				break;
			default:
				exit(1);
		}
	}

	/*
	 * If the incorrect number of arguments have been given then display
	 * the correct program usage and exit.
	 */
	if((argc - optind) < 4) {
		usage(argv[0]);
		exit(1);
	}

	/*
	 * If the user is logging the output to a file then open the logfile.
	 */
	if(log_path) {
		printf("[*] Opening %s for writing: ", log_path);
		if((log_fd = fopen(log_path, "w")) == NULL) {
			printf("failure\n");
			exit(1);
		} else
			printf("success\n\n");
	}

	/*
	 * If the target host is being resolved then resolve it and place the
	 * numerical IP address in the host variable of the packet structure,
	 * if it is not then place the hostname in the host variable of the
	 * packet structure.
	 */
	if(packet.resolved == PACKET_RESOLVED_YES) {
		printf("[*] Resolving %s: ", argv[optind]);
		if((ip_address = get_ip(argv[optind])) == -1) {
			printf("failure\n");
			if(log_path)
				fclose(log_fd);
			exit(1);
		} else {
			printf("success\n\n");
#ifndef NO_SNPRINTF
			snprintf(packet.host, LENGTH_HOST - 1,
			  "%ld.%ld.%ld.%ld", ip_address & 255, ip_address >> 8
			  & 255, ip_address >> 16 & 255, ip_address >> 24 &
			  255);
#else
			sprintf(packet.host, "%ld.%ld.%ld.%ld", ip_address &
			  255, ip_address >> 8 & 255, ip_address >> 16 & 255,
			  ip_address >> 24 & 255);
#endif
		}
	} else
		strncpy(packet.host, argv[optind], LENGTH_HOST - 1);

	/*
	 * Copy the list of distributed hosts to the linked list.
	 */
	if(!suppress)
		printf("[*] Adding distributed hosts to list: ");
	
	DHOST_LIST_INIT(list);
	for(i = optind + 3; i < argc; i++)
		ADD_DHOST(list, argv[i], DSCAND_PORT);

	if(!suppress)
		printf("success\n");

	/*
	 * Resolve the IP addresses of the distributed hosts and check that
	 * they are listening on port 17001.
	 */
	if(!suppress)
		printf("[*] Resolving and connecting to distributed hosts:\n" \
		  "\n    [ Host                  ] [ Numerical IP address ]\n");

	for(current_dhost = list.head; current_dhost; current_dhost =
	  current_dhost->next) {
	  	if((ip_address = get_ip(current_dhost->hostname)) == -1) {
	  		if(!suppress)
	  			printf("      %21.21s     Failure resolving\n",
	  		  	  current_dhost->hostname);
	  		DEL_DHOST(list, current_dhost->hostname);
	  	} else {
	  		current_dhost->addr = ip_address;

			if((current_dhost->sock = socket(AF_INET, SOCK_STREAM,
			  0)) < 0) {
			  	FATAL_ERROR("cannot create socket");
				if(log_path)
					fclose(log_fd);
				exit(1);
			}

			sock_info.sin_port = htons(current_dhost->port);
			sock_info.sin_family = AF_INET;
			sock_info.sin_addr.s_addr = current_dhost->addr;

			if(connect(current_dhost->sock, (struct sockaddr *)
			  &sock_info, sizeof(sock_info)) < 0) {
				if(!suppress)
					printf("      %21.21s     Failure "   \
					  "connecting\n",
					  current_dhost->hostname);
				DEL_DHOST(list, current_dhost->hostname);
				continue;
			}

			close(current_dhost->sock);
	  		if(!suppress)
	  			printf("      %21.21s     %ld.%ld.%ld.%ld\n",
	  			  current_dhost->hostname, current_dhost->addr
	  			  & 255, current_dhost->addr >> 8 & 255,
	  			  current_dhost->addr >> 16 & 255,
	  			  current_dhost->addr >> 24 & 255);
	  	}
	}

	/*
	 * If the results are being logged then dump the table header to the
	 * log file.
	 */
	if(log_path)
		fprintf(log_fd, "[*] Scanning %s:\n\n    [ Port  ] [ Scanned "\
		  "      ] [ Status         ]\n", argv[optind]);

	/*
	 * Scan the target host using a different distributed host each
	 * time.
	 */
	current_dhost = list.head;
	printf("\n[*] Scanning %s:\n\n    [ Port  ] [ Scanned by    ] [ " \
	  "Status         ]\n", argv[optind]);
	for(packet.port = atoi(argv[optind + 1]); packet.port <=
	  atoi(argv[optind + 2]); packet.port++) {
	  	/*
	  	 * Check that there are still some distributed hosts left
	  	 * in the list, if there are not then display a fatal error
	  	 * and exit.
	  	 */
	  	if(list.num_dhosts <= 0) {
	  		FATAL_ERROR("no distributed hosts left in list");
			if(log_path)
				fclose(log_fd);
			exit(1);
	  	}

		/*
		 * Connect to the distributed host.
		 *
		 * XXX - In the next release the pool of sockets will be opened
		 * at the start and will not be closed again therefore cutting
		 * down on the number of calls to socket() and connect().
		 */
		if((current_dhost->sock = socket(AF_INET, SOCK_STREAM,
		  0)) < 0) {
			FATAL_ERROR("cannot create socket");
			if(log_path)
				fclose(log_fd);
			exit(1);
		}

		sock_info.sin_port = htons(current_dhost->port);
		sock_info.sin_family = AF_INET;
		sock_info.sin_addr.s_addr = current_dhost->addr;

		if(connect(current_dhost->sock, (struct sockaddr *) &sock_info,
		  sizeof(sock_info)) < 0) {
			printf("[*] Error: failure connecting to %s\n",
			  current_dhost->hostname);
			DEL_DHOST(list, current_dhost->hostname);
			packet.port--;
			continue;
		}

		/*
		 * Send the packet to the distributed host and recieve the
		 * reply.
		 */
		packet.type = PACKET_TYPE_REQUEST;
		packet.reply = PACKET_REPLY_NULL;

		bzero(&packet_buffer, sizeof(packet_buffer));
#ifndef NO_SNPRINTF
		CONSTRUCT_PACKET(packet, packet_buffer);
#else
		CONSTRUCT_PACKET_SPRINTF(packet, packet_buffer);
#endif

		if(write(current_dhost->sock, &packet_buffer,
		  sizeof(packet_buffer)) < 0) {
		  	printf("[*] Error: write failure\n");
		  	DEL_DHOST(list, current_dhost->hostname);
		  	packet.port--;
			continue;
		}

		bzero(&packet_buffer, sizeof(packet_buffer));
		if(read(current_dhost->sock, &packet_buffer,
		  sizeof(packet_buffer)) < 0) {
			printf("[*] Error: read failure\n");
			DEL_DHOST(list, current_dhost->hostname);
			packet.port--;
			continue;
		}

		if(DECONSTRUCT_PACKET(recv_packet, packet_buffer) <
		  DECONSTRUCT_PACKET_SUCCESS) {
			printf("[*] Error: DECONSTRUCT_PACKET failure\n");
			DEL_DHOST(list, current_dhost->hostname);
			packet.port--;
			continue;
		}

		close(current_dhost->sock);

		/*
		 * Display the results of the scan.
		 */
		switch(recv_packet.reply) {
		case PACKET_REPLY_OPEN:
			printf("      %5d     %13.13s     Open\n", packet.port,
			  current_dhost->hostname);
			if(log_path) 
				fprintf(log_fd, "      %5d     %13.13s" \
				  "     Open\n", packet.port,
				  current_dhost->hostname);
			break;
		case PACKET_REPLY_CLOSED:
			/* we are no longer going to display closed ports */
			break;
		case PACKET_REPLY_FAILURE:
			printf("      %5d     %13.13      Internal dscand " \
			  "failure\n", packet.port, current_dhost->hostname);
			break;
		case PACKET_REPLY_UNIMP:
			printf("      %5d     %13.13      Currently " \
			  "unimplemented\n", packet.port,
			  current_dhost->hostname);
			break;
		default:
			printf("      %5d     %13.13      Unknown reply " \
			  "value\n", packet.port, current_dhost->hostname);
			break;
		}

	  	/*
	  	 * Point the distributed host to the next one in the list,
	  	 * if the end of the list has been reached point back to the
	  	 * start of the list.
	  	 */
	  	if(!current_dhost->next)
	  		current_dhost = list.head;
	  	else
	  		current_dhost = current_dhost->next;

	  	/*
	  	 * If the user has specified a sleep time then sleep for that
	  	 * length of time before moving onto the next port.
	  	 */
	  	if(sleep_time != 0)
	  		sleep(sleep_time);
	}

	/*
	 * If the user is logging the output to a file then close the logfile.
	 */
	if(log_path) {
		printf("\n[*] Closing %s: ", log_path);
		if(fclose(log_fd) != 0) {
			printf("failure\n");
			exit(1);
		} else
			printf("success\n");
	}

	exit(0);
}

void	usage(char *name)
{
	fprintf(stderr, "Usage: %s [-hvuRS] [-p sec] [-l file] host "         \
	  "start_port stop_port dhosts...\n", name);
	fprintf(stderr, "  -h          display correct program usage\n");
	fprintf(stderr, "  -v          display program version\n");
	fprintf(stderr, "  -u          scan UDP ports, default is TCP\n");
	fprintf(stderr, "  -R          make the distributed host resolve the "\
	  "hostname\n");
	fprintf(stderr, "  -S          suppress output\n");
	fprintf(stderr, "  -p sec      pause for sec seconds before scanning "\
	  "next port\n");
	fprintf(stderr, "  -l file     log output to file\n");
	fprintf(stderr, "  host        host to scan\n");
	fprintf(stderr, "  start_port  port to start scanning from\n");
	fprintf(stderr, "  stop_port   port to stop scanning at\n");
	fprintf(stderr, "  dhosts...   list of distributed hosts\n");
}
