/*++
/* NAME
/*	dscand 8
/* SUMMARY
/*	distributed port scanner daemon
/* SYNOPSYS
/* .fi
/*	\fBdscandf\fR [\fB-hv\fR] \fIcommand\fR
/* DESCRIPTION
/* 	\fBdscand\fR scans ports on computers under instruction from
/*	\fBdscan\fR.  The following options are implemented:
/*
/* .IP \fB-h\fR
/*	Display the correct program usage.
/* .IP \fB-v\fR
/*	Display the program version.
/* .PP
/*	The following commands are implemented:
/* .IP \fBstart\fR
/*	Start the distributed port scanner daemon.
/* .IP \fBstop\fR
/*	Stop the distributed port scanner daemon.
/* .IP \fBsimple\fR
/*	Starts the distributed port scanner without saving the current
/* 	pid in a file.
/* SEE ALSO
/*	dscan(1) distributed port scanner
/* LICENSE
/* .ad
/* .fi
/*	The dscan license must be distributed with this software.
/* AUTHOR(S)
/*	Intrinsic <intrinsic@hushmail.com>, Ni9 Technologies
/*--*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <dscan.h>
#include <pid_file.h>
#include <network.h>

void	usage(void);
void	dscand_stop(void);

int	sock;
char	*name;

int	main(int argc, char **argv)
{
	int	arg, nsock;
	char	packet_buffer[LENGTH_PACKET];
	pid_t	pid;
	struct sockaddr_in sock_info;
	struct dscanp_packet_s packet;

	/*
	 * Create a global pointer to argv[0] for use in error messages.
	 */
	name = argv[0];

	/*
	 * Parse any flags given.
	 */
	while((arg = getopt(argc, argv, "hv")) != EOF) {
		switch(arg) {
			case 'h':
				usage();
				exit(0);
			case 'v':
				fprintf(stderr, "%s\n", VERSION);
				exit(0);
			default:
				exit(1);
		}
	}

	/*
	 * Check that there has been a command given before parsing it.
	 */
	if((argc - optind) < 1) {
		usage();
		exit(1);
	}
	
	if(strcmp(argv[optind], "start") == 0) {
		/*
		 * Check that the distributed port scanner daemon has not
		 * already been started.
		 */
		if(pid_file_check(PID_FILE) >= 0) {
			fprintf(stderr, "%s: process already started\n", name);
			exit(1);
		}

		/*
		 * Write the current pid to the pid file.
		 */
		if(pid_file_write(PID_FILE) < 0) {
			fprintf(stderr, "%s: %s: failure writing pid to file "\
			  "\n", name, PID_FILE);
			exit(1);
		}

		/*
		 * Set up the signal handlers.
		 */
		signal(SIGINT, (void *) dscand_stop);
	} else if(strcmp(argv[optind], "simple") == 0) {
		/*
		 * Tell the user how to stop the daemon.
		 */
		printf("To stop the daemon send a SIGINT signal to %d\n",
		  getpid());
	} else if(strcmp(argv[optind], "stop") == 0) {
		/*
		 * Read the pid of the daemon process from the pid file.
		 */
		if((pid = pid_file_check(PID_FILE)) < 0) {
			fprintf(stderr, "%s: could not find process\n", name);
			exit(1);
		}

		/*
		 * Send the SIGINT signal to the process which should cause it
		 * to exit.
		 */
		if(kill(pid, SIGINT) < 0) {
			fprintf(stderr, "%s: failure sending signal to "      \
			  "process\n", name);
			exit(1);
		}

		exit(0);
	} else {
		fprintf(stderr, "%s: %s: unknown command\n", name, 
		  argv[optind]);
		exit(1);
	}

	/*
	 * Create a socket and bind it to a port.
	 */
	if((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		fprintf(stderr, "%s: failure creating socket\n", name);
		dscand_stop();
	}

	sock_info.sin_family = AF_INET;
	sock_info.sin_port = htons(DSCAND_PORT);
	sock_info.sin_addr.s_addr = INADDR_ANY;

	if(bind(sock, (struct sockaddr *) &sock_info, sizeof(sock_info)) < 0) {
		fprintf(stderr, "%s: %d: bind failure\n", name, DSCAND_PORT);
		dscand_stop();
	}

	if(listen(sock, 5) < 0) {
		fprintf(stderr, "%s: listen failure\n", name);
		dscand_stop();
	}

	/*
	 * Enter an infinate loop waiting connetions.
	 */
	while(1) {
		if((nsock = accept(sock, 0, 0)) < 0) 
			continue;

		/*
		 * Fork to handle communications with the client.
		 */
		if(fork() == 0) {
			close(sock);

			/*
			 * Get the packet and deconstruct it.
			 */
			bzero(&packet_buffer, sizeof(packet_buffer));
			if(read(nsock, &packet_buffer, sizeof(packet_buffer))
			  < 0) {
				close(nsock);
				exit(1);
			}

			if(DECONSTRUCT_PACKET(packet, packet_buffer) <
			  DECONSTRUCT_PACKET_SUCCESS) {
				close(nsock);
				exit(1);
			}

			/*
			 * Scan the port and send the results to the client.
			 */
			packet.type = PACKET_TYPE_REPLY;
			packet.reply = scan_port(packet);

			bzero(&packet_buffer, sizeof(packet_buffer));
#ifndef NO_SNPRINTF
			CONSTRUCT_PACKET(packet, packet_buffer);
#else
			CONSTRUCT_PACKET_SPRINTF(packet, packet_buffer);
#endif

			if(write(nsock, &packet_buffer, sizeof(packet_buffer))
			  < 0) {
				close(nsock);
				exit(1);
			}

			close(nsock);
			exit(0);
		}

		close(nsock);
	}
}

void	usage(void)
{
	fprintf(stderr, "Usage: %s [-hv] command\n", name);
	fprintf(stderr, "  -h        display correct program usage\n");
	fprintf(stderr, "  -v        display program version\n");
	fprintf(stderr, "  command   start, stop or simple\n");
}

void	dscand_stop(void)
{
	/*
	 * Remove the pid file, if this fails it does not really matter as the
	 * daemon may have been started in simple mode and therefore a pid file
	 * would not exist.
	 */
	pid_file_delete(PID_FILE);

	close(sock);

	exit(0);
}

