/*++
/* NAME
/*	linked_list 3
/* SUMMARY
/*	module for manipulating a linked list of distributed hosts
/* SYNOPSYS
/*	#include <linked_list.h>
/*
/*	int	add_dhost(list, hostname, port)
/*	struct dhost_list_s *list;
/*	char	*hostname;
/*	unsigned int port;
/*
/*	int	del_dhost(list, hostname)
/*	struct dhost_list_s *list;
/*	char	*hostname;
/* DESCRIPTION
/*	add_dhost() adds a distributed host to a list, returns -1 if the
/*	memory needed could not be allocated and 0 on success.
/*
/*	del_dhost() removes a distributed host from a list, returns -1 if
/*	the distributed host could not be found and 0 on success.
/* DIAGNOSTICS
/* LICENSE
/* .ad
/* .fi
/*	The dscan license msut be distributed with this software.
/* AUTHOR(S)
/*	Intrinsic <intrinsic@hushmail.com>, Ni9 Technologies
/*--*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <linked_list.h>

int	add_dhost(struct dhost_list_s *list, char *hostname, unsigned int
	  port)
{
	struct dhost_s *new_dhost = NULL;

	/*
	 * Allocate the memory for the new list element, if this fails then
	 * return -1.
	 */
	if(!(new_dhost = (struct dhost_s *) malloc(sizeof(struct dhost_s))))
		return(-1);

	/*
	 * Allocate the memory for the hostname string inside the list element
	 * and copy the hostname given into it, then copy the other variables
	 * into the list.  On memory allocation failure return -1.
	 */
	if(!(new_dhost->hostname = (char *) malloc(sizeof(hostname) + 1)))
		return(-1);
	strcpy(new_dhost->hostname, hostname);
	new_dhost->addr = 0;
	new_dhost->port = port;
	new_dhost->next = NULL;
	new_dhost->prev = NULL;

	/*
	 * Add the new element to the list.
	 */
	if(!list->head) {
		list->head = new_dhost;
		list->tail = new_dhost;
	} else {
		list->tail->next = new_dhost;
		new_dhost->prev = list->tail;
		list->tail = new_dhost;
	}

	list->num_dhosts++;
	return(0);
}

int	del_dhost(struct dhost_list_s *list, char *hostname)
{
	struct dhost_s *current_dhost = NULL;

	/*
	 * Locate the distributed host in the list, if it is not present
	 * return -1.
	 */
	for(current_dhost = list->head; current_dhost; current_dhost =
	  current_dhost->next) {
		if(strcmp(current_dhost->hostname, hostname) == 0)
			break;
	}

	if(current_dhost == NULL)
		return(-1);

	/*
	 * Remove the server, if DONT_FREE is defined then the free() call
	 * is not used and the element is not actualy removed from memory
	 * rather it is orphaned.
	 */
	if(current_dhost->next && current_dhost->prev) {
		current_dhost->prev->next = current_dhost->next;
		current_dhost->next->prev = current_dhost->prev;
#ifndef DONT_FREE
		free(current_dhost);
#endif
	} else if(current_dhost->next) {
		current_dhost->next->prev = NULL;
		list->head = current_dhost->next;
#ifndef DONT_FREE
		free(current_dhost);
#endif
	} else if(current_dhost->prev) {
		current_dhost->prev->next = NULL;
		list->tail = current_dhost->prev;
#ifndef DONT_FREE
		free(current_dhost);
#endif
	} else {
		list->head = NULL;
		list->tail = NULL;
#ifndef DONT_FREE
		free(current_dhost);
#endif
	}

	list->num_dhosts--;
	return(0);
}
